package robots;

import java.io.*;
import java.util.*;

/**
 * Trieda realizujuca samotnu simulaciu robota na toruse s polickami, na ktorych mozu byt ulozene prvky typu T.
 */
public class Simulation {
    private int height;
    private int width;
    private List<List<Boolean>> torus;
    private Robot robot;
    private int row;
    private int column;

    /**
     * Konstruktor, ktory vytvori simulaciu na zaklade nasledujucich parametrov:
     * - Dvojrozmerneho zoznamu torus reprezentujuceho torus s polickami obsahujucimi zaznamy typu T. Musi ist
     *   o obdlznikovy zoznam nenulovej sirky a vysky -- inak dojde k vyhodeniu vynimky typu InvalidTorusException.
     * - Robota robot, ktoreho pohyb na toruse sa bude simulovat.
     * - Celych cisel initialRow a initialColumn udavajucich pociatocnu poziciu robota robot na toruse. Pociatocna
     *   pozicia musi byt v rozsahu zoznamu torus -- inak dojde k vyhodeniu vynimky typu InvalidRobotPositionException.
     */
    public Simulation(List<List<Boolean>> torus, Robot robot, int initialRow, int initialColumn) {
        if (torus == null) {
            throw new InvalidTorusException();
        }
        height = torus.size();
        if (height <= 0) {
            throw new InvalidTorusException();
        }
        width = torus.get(0).size();
        if (width <= 0) {
            throw new InvalidTorusException();
        }
        this.torus = new ArrayList<>();
        for (List<Boolean> row : torus) {
            if (row.size() != width) {
                throw new InvalidTorusException();
            }
            this.torus.add(new ArrayList<>(row));
        }
        this.robot = robot;
        if (initialRow < 0 || initialRow >= height || initialColumn < 0 || initialColumn >= width) {
            throw new InvalidRobotPositionException();
        }
        row = initialRow;
        column = initialColumn;
    }

    private Boolean currentCellContents() {
        return torus.get(row).get(column);
    }

    /**
     * Metoda, ktora spusti nanajvys maxSteps krokov simulacie vytvorenej konstruktorom. Ak maxSteps < 0, pobezi
     * simulacia az kym sa robot zastavi.
     *
     * V pripade, ze outputMode.equals("steps"), bude sa do vystupneho prudu out vypisovat postupnost krokov
     * vykonavanych robotom.
     * V pripade, ze outputMode.equals("cells"), vypise sa po ukonceni simulacie do vystupneho prudu out obsah
     * jednotlivych policok torusu.
     */
    public void start(PrintStream out, int maxSteps, String outputMode) {
        int steps = 0;
        while ((steps < maxSteps || maxSteps < 0) && robot.hasNextMove(currentCellContents())) {
            steps++;
            Move move = robot.nextMove(currentCellContents());
            torus.get(row).set(column, move.getNewCellContents());
            switch (move.getDirection()) {
                case LEFT:
                    column = Math.floorMod(column - 1, width);
                    break;
                case UP:
                    row = Math.floorMod(row - 1, height);
                    break;
                case RIGHT:
                    column = Math.floorMod(column + 1, width);
                    break;
                case DOWN:
                    row = Math.floorMod(row + 1, height);
                    break;
            }
            if (outputMode.equals("steps")) {
                out.println(move.getNewCellContents() + " " + move.getDirection());
            }
        }
        if (outputMode.equals("cells")) {
            for (int i = 0; i <= height - 1; i++) {
                for (int j = 0; j <= width - 1; j++) {
                    if (torus.get(i).get(j)) {
                        out.print("T");
                    } else {
                        out.print("F");
                    }
                }
                out.println();
            }
        }
    }

}
